/*
 *
 * B2BUAEndPoint.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: B2BUAEndPoint.cxx,v $
 * Revision 1.22  2007/07/14 07:50:29  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.21  2007/07/03 17:17:37  joegenbaclor
 * Got rid of compile warnings in unix builds
 *
 * Revision 1.20  2007/06/28 15:47:55  joegenbaclor
 * Enhanced buggy subscribe, notify, publish B2BUA pass through
 *
 * Revision 1.19  2007/06/19 11:25:13  joegenbaclor
 * Allowed Subscribe, Notify and Publish to be proxied by the B2BUA Endpoint.  We did this by introducing custom ProcessStackEvent handlers for B2BUAEndPoint
 *
 * Revision 1.18  2007/05/25 08:38:09  joegenbaclor
 * Introduced GCVERIFYREF and GCCREATEREF Macro
 * Transfered B2BUAConnection::OnIncomingCall to be called in B2BUAEndPoint::OnIncomingSIPMessage
 *
 * Revision 1.17  2007/05/01 04:20:32  rcolobong
 * 1. Add OnCallRedirect method for handling Local Redirect
 * 2. Add support for enabling and disabling Local Refer
 *
 * Revision 1.16  2007/04/17 08:02:15  joegenbaclor
 * more work on local refer
 *
 * Revision 1.15  2007/04/16 14:24:41  joegenbaclor
 * More work on B2B local REFER support
 *
 * Revision 1.14  2007/03/28 09:09:24  joegenbaclor
 * Logging enhancements
 *
 * Revision 1.13  2007/02/09 03:47:35  joegenbaclor
 * 1. Added MaxForwards handling for B2B calls
 * 2. #defined SIP_DEFAULT_MAX_FORWARDS constant
 *
 * Revision 1.12  2007/01/14 13:12:54  joegenbaclor
 * Fixed compile warnings
 *
 * Revision 1.11  2007/01/11 02:11:41  rcolobong
 * Add setter for enable/disable proxying media
 *
 * Revision 1.10  2006/11/30 08:19:37  joegenbaclor
 * Relicensed opensipstack to MPL/GPL/LGPL triple license scheme
 *
 * Revision 1.9  2006/11/23 11:00:12  joegenbaclor
 * Added Instant Message B2B support
 *
 * Revision 1.8  2006/11/10 07:58:05  rcolobong
 * Add Method to disable/enable Session Timers in B2BUACall
 *
 * Revision 1.7  2006/08/30 03:42:05  joegenbaclor
 * Corrected bug in garbage collector loop.
 *
 * Revision 1.6  2006/07/19 02:19:21  joegenbaclor
 * 1.  Bug fixes concerning B2BUA destruction during during 3xx, 4xx, 5xx, 6xx local response
 * 2.  Fixed bug where Route header is never removed from outbound Invite resulting that may result to a loop when
 *      Inbound Invite has a Route header present
 * 3.  Added RemoveAllRoutes() in SIPMessage
 *
 * Revision 1.5  2006/07/10 06:29:38  joegenbaclor
 * Completed crude Registration support for B2BUA
 *
 * Revision 1.4  2006/06/20 09:52:55  joegenbaclor
 * Added Routing and SDP Handlers to B2BUA
 *
 * Revision 1.3  2006/06/14 08:43:38  joegenbaclor
 * Initial upload of OpenB2BUA applcation and related classes
 *
 * Revision 1.2  2006/04/10 01:09:03  joegenbaclor
 * Implemented Create methods fo B2BUA sessions
 *
 * Revision 1.1  2006/04/08 06:27:13  joegenbaclor
 * Initial upload of B2BUA classes
 *
 *
 */

#include "B2BUAEndPoint.h"
#include "B2BUAConnection.h"

#define new PNEW

using namespace B2BUA;

B2BUAEndPoint::B2BUAEndPoint(
  SIPUserAgent & ua,
  PINDEX sessionThreadCount,
  PINDEX stackSize
) : CallSessionManager( ua, sessionThreadCount, stackSize )
{
  m_EnableSessionTimer = FALSE;
  m_MediaProxyIfPrivate = FALSE;
  m_EnableLocalRefer = FALSE;
}

B2BUAEndPoint::~B2BUAEndPoint()
{
}

B2BUAConnection * B2BUAEndPoint::CreateB2BUA(
  const SIPMessage & request,
  B2BUACall * call
)
{
  PWaitAndSignal lock( m_SessionsMutex );

  OString sessionId = call->GetSessionId();
  OString b2buaId = sessionId + OString( "-connection" );

  B2BUAConnection * b2bua = OnCreateB2BUA( 
    request, 
    b2buaId, 
    call 
  );

  if( b2bua == NULL )
    return NULL;

  b2bua->SetLogContextId( sessionId );

  b2bua->AttachCall( sessionId, call );

  b2bua->EnableLocalRefer( m_EnableLocalRefer );
   
  m_SessionsBySessionId.SetAt( b2buaId.c_str(), b2bua );
  m_SessionsByCallId.SetAt( b2buaId.c_str(), b2bua );

  return b2bua;
}

B2BUAConnection * B2BUAEndPoint::OnCreateB2BUA(
  const SIPMessage & /*request*/,
  const OString & sessionId,
  B2BUACall * /*call*/
)
{
  B2BUAConnection * connection = new B2BUAConnection( *this, sessionId );
  connection->SetMediaProxyIfPrivate( m_MediaProxyIfPrivate );
  return connection;
}

void B2BUAEndPoint::ProcessStackEvent(
  SIPSessionManager * proxy,
  SIPStackEvent * eventObject
)
{
  SIPSession::GCRef gcRef = FindGCRefByCallId( eventObject->GetCallId() );
  ProcessStackEvent( proxy, gcRef, eventObject );
}

void B2BUAEndPoint::ProcessStackEvent(
  SIPSessionManager * proxy,
  SIPSession::GCRef & gcRef,
  SIPStackEvent * event 
)
{
  if( gcRef == NULL )
  {
    /// there is no session for this message
    /// check if this is a subscribe, notify or publish
    TransactionId transactionId = event->GetTransactionId();
    OString method = transactionId.GetMethod().ToUpper();
    OString callId = transactionId.GetCallId();
    To to;
    OString domain;
    SIPMessage msg;

    if( event->GetType() == SIPStackEvent::Message )
    {
      SIPMessageArrival & msgEvent = (SIPMessageArrival&)*event;
      msg = msgEvent.GetMessage();

      if( msg.IsSubscribe() || msg.IsNotify() || msg.IsPublish() )
      {
        SIPURI rURI;
        if( msg.GetRequestURI( rURI ) )
        {
          if( !GetUserAgent().GetStack().GetTransportManager()->IsLocalAddressAndPort( rURI ) )
          {
            proxy->ProcessStackEvent( gcRef, event );
            return;
          }
        }

        SIPURI toURI = msg.GetToURI();
        if( !GetUserAgent().GetStack().GetTransportManager()->IsLocalAddressAndPort( toURI ) )
        {
          proxy->ProcessStackEvent( gcRef, event );
          return;
        }
      }else if( !msg.IsRequest() )
      {
        OString meth = msg.GetCSeqMethod();
        if( meth == "SUBSCRIBE" ||
            meth == "NOTIFY" ||
            meth == "PUBLISH" )
        {
          ///this is a response
          SIPSession::GCRef ref = proxy->FindGCRefByCallId( msg.GetCallId() );
          if( ref != NULL )
          {
            proxy->ProcessStackEvent( ref, event );
            return;
          }
        }
      }
    }
  }
  static_cast<SIPSessionManager *>(this)->ProcessStackEvent( gcRef, event );
}

void B2BUAEndPoint::OnIncomingSIPMessage(
  SIPMessageArrival & messageEvent,
  SIPSession * session
)
{
  const TransactionId & transactionId = messageEvent.GetTransactionId();
  OString callId = transactionId.GetCallId();
  SIPMessage request = messageEvent.GetMessage();

  if( !request.IsValid() )
    return;

  if( session != NULL )
  {

    LOG_IF_DEBUG( session->LogDetail(), "*** MESSAGE ARRIVAL *** for SIP Session " << session->GetSessionId() );
    LOG_IF_DEBUG( session->LogDebugVeryHigh(), request );
    session->OnIncomingSIPMessage( messageEvent );
    return;
  }else
  {
    if( request.IsRequest() )
    {
      if( !request.IsAck())
      {
        SIPSession::GCRef ref = CreateServerSession( request );
        session = ref.GetObject();

        if( session != NULL )
        {
          if( dynamic_cast<B2BUACall*>(session)->GetB2BUAConnection().OnIncomingCall( dynamic_cast<B2BUACall&>(*session), request ) )
          {
            LOG( session->LogDetail(), "*** MESSAGE ARRIVAL *** for SIP Session " << session->GetSessionId() );
            LOG_IF_DEBUG( session->LogDebugVeryHigh(), request );
            SIPMessageArrival messageEvent2( request, messageEvent.GetTransactionId() );
            session->OnIncomingSIPMessage( messageEvent2 );
          }
          return;
        }else
        {
          LOG( LogDetail(), "*** MESSAGE ARRIVAL *** No Session available to handle " << request.GetStartLine() );
          LOG_IF_DEBUG( LogDebugVeryHigh(), request );
          OnOrphanedMessage( request );
        }
      }else
      {
        OnOrphanedMessage( request );
      }

    }else  /// this is a response 
    {
      LOG_CONTEXT_IF_DEBUG( LogWarning(), request.GetCallId(), "*** Received a response for a deleted session *** "  );
      LOG_CONTEXT_IF_DEBUG( LogDebugVeryHigh(), request.GetCallId(), request );
      OnOrphanedMessage( request );
      return;
    }
  }
  
}

CallSession * B2BUAEndPoint::OnCreateServerCallSession(
  SIPMessage & request,
  const OString & sessionId
)
{
  if( !request.IsInvite() )
    return NULL;

  /// incoming INVITE... create a session
  B2BUACall * call = new B2BUACall( *this, request, sessionId );
  call->SetLegIndex( 0 );
  B2BUAConnection * b2bua = CreateB2BUA( request, call );

  if( b2bua == NULL )
  {
    delete call;
    return NULL;
  }

  call->SetB2BUAConnection( b2bua );

  return call;
}


CallSession * B2BUAEndPoint::OnCreateClientCallSession(
  const ProfileUA & profile,
  const OString & sessionId
)
{
  B2BUACall * call = new B2BUACall( *this, sessionId, profile );
  call->SetLegIndex( 1 );
  call->EnableSessionTimer( m_EnableSessionTimer );

  return call;
}

void B2BUAEndPoint::OnReceivedInstantMessage(
  const SIPMessage & message
)
{
  B2BUserAgent& ua = dynamic_cast<B2BUserAgent&>( GetUserAgent() );
  ua.OnReceivedInstantMessage( message );
}

BOOL B2BUAEndPoint::OnCallRedirect(
  CallSession & session,
  SIPMessage & redirect
)
{
  B2BUserAgent& ua = dynamic_cast<B2BUserAgent&>( GetUserAgent() );
  B2BUACall & call = dynamic_cast<B2BUACall&>( session );

  return ua.OnCallRedirect( call.GetB2BUAConnection(), call, redirect );
}

