/*
 *
 * MediaServer.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: MediaServer.cxx,v $
 * Revision 1.8  2007/07/14 07:50:32  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.7  2007/06/05 16:27:56  joegenbaclor
 * Added IVR audio playback functions to B2BUAConnection
 *
 * Revision 1.6  2007/06/04 16:05:33  joegenbaclor
 * *** empty log message ***
 *
 * Revision 1.5  2007/05/29 17:42:33  joegenbaclor
 * Initial work on auto attendant
 *
 * Revision 1.4  2007/03/08 10:23:39  joegenbaclor
 * Tidy up OpalManager Destruction
 * Enabled Recording in vxml
 *
 * Revision 1.3  2007/03/08 06:43:18  joegenbaclor
 * This commit marks first successful call to the Media Server
 *
 * Revision 1.2  2007/03/08 03:36:47  joegenbaclor
 * Transferred declarration of allcodecs.h in transacoders.cxx to avoid multiple definition
 *
 * Revision 1.1  2007/03/07 23:20:22  joegenbaclor
 * Added media server files
 *
 *
 */

#include "MediaServer.h"
#include "MediaServerSession.h"
#include "B2BUAConnection.h"

#include <opal/transcoders.h>
#define new PNEW

using namespace MS;
using namespace B2BUA;

MediaServer::MediaServer( 
  CallSessionManager * b2bUA 
)
{
  m_B2BUA = b2bUA;
  m_MediaServerEP = NULL;
  m_SIPEP = NULL;
  m_InterfaceAddress = "*";
  m_ListenerPortBase = m_ListenerPort = 65070;
  m_IsInitialized = FALSE;
}

MediaServer::~MediaServer()
{
  ClearAllCalls();
  // Shut down the cleaner thread
  garbageCollectExit.Signal();
  garbageCollector->WaitForTermination();

  if( m_MediaServerEP != NULL )
    delete m_MediaServerEP;

  if( m_SIPEP != NULL )
    delete m_SIPEP;
}

CallSessionManager * MediaServer::GetB2BUA()const
{ 
  return m_B2BUA; 
}

BOOL MediaServer::Initialize()
{
  if( m_IsInitialized )
    return TRUE;

  LOG( LogInfo(), "Media Server SIP Stack Initialization Started" );

  if( m_MediaServerEP == NULL )
    m_MediaServerEP = CreateMediaServerEP();

#ifdef WIN32
  m_MediaServerEP->SetDefaultTextToSpeech( "Microsoft SAPI" );
#endif

  if( m_SIPEP == NULL )
    m_SIPEP = CreateSIPEP();

  AddRouteEntry("ivr:.*  = sip:<da>");
  AddRouteEntry("sip:.*  = ivr:<da>");

  InitMediaInfo(m_MediaServerEP->GetPrefixName(), m_MediaServerEP->GetMediaFormats());
  OpalMediaFormatList mediaFormats;
  mediaFormats += m_MediaServerEP->GetMediaFormats();
  InitMediaInfo( "sw", OpalTranscoder::GetPossibleFormats( mediaFormats ) );
  ApplyMediaInfo();

  /// determine the port to use
  WORD sipPort = m_ListenerPortBase;
  {
    ///do a dummy listen to check the available port
    for( PINDEX i = sipPort; i < 10000; i++ )
    {
      PUDPSocket sock( sipPort );

      if( sock.Listen( PIPSocket::Address( m_InterfaceAddress == "*" ? "127.0.0.1" : m_InterfaceAddress ) ) )
      {
        sock.Close();
        break;
      }
      ++sipPort;
    }
  }
  
  PIPSocket::Address udpAddr (m_InterfaceAddress.c_str());
  m_SIPEP->GetProfile().GetTransportProfile().EnableUDP( udpAddr, sipPort );

#if ENABLE_TCP_TRANSPORT
    m_SIPEP->GetProfile().GetTransportProfile().EnableTCP( m_InterfaceAddress, sipPort);
#endif

  m_SIPEP->Initialize();

  m_ListenerAddress = m_SIPEP->GetUserAgent()->GetStack().GetTransportManager()->GetDefaultInterfaceAddress();
  m_ListenerPort = sipPort;

  return m_IsInitialized = TRUE;
}

MediaServerEndPoint * MediaServer::CreateMediaServerEP()
{
  return new MediaServerEndPoint( *this );
}

OpalOSSEndPoint * MediaServer::CreateSIPEP()
{
  return new OpalOSSEndPoint( *this );
}

OpalIVRConnection * MediaServer::CreateConnection(
  OpalCall & call,
  const PString & token,
  void * userData,
  const PString & vxml
)
{
  OString callId = call.GetSessionIdentifier();
  B2BUAEndPoint * b2bua = dynamic_cast<B2BUAEndPoint *>( m_B2BUA );
  SIPSession::GCRef callRef = b2bua->FindGCRefByCallId( callId );
  
  if( callRef == NULL )
    return NULL;

  B2BUACall * b2bUACall = dynamic_cast<B2BUACall *>(callRef.GetObject());
  if( b2bUACall == NULL )
    return NULL;

  MediaServerSession * conn = new MediaServerSession(
    call,            ///  Owner calll for connection
    *m_MediaServerEP, ///  Owner endpoint for connection
    token      ///  Token for connection
  );

  IVRContext * context = new IVRContext(conn, PFactory<PTextToSpeech>::CreateInstance(m_MediaServerEP->GetDefaultTextToSpeech()));

  conn->SetIVRContext( context );
  b2bUACall->GetB2BUAConnection().AttachIVRContext( context );
  
  return conn;
}

void MediaServer::InitMediaInfo(const char * source, const OpalMediaFormatList & mediaFormats)
{
  for (PINDEX i = 0; i < mediaFormats.GetSize(); i++) {
    const OpalMediaFormat & mediaFormat = mediaFormats[i];
    if (mediaFormat.GetPayloadType() != RTP_DataFrame::MaxPayloadType)
    {
      m_MediaInfo.push_back(CodecList(source, mediaFormat));
      m_MediaFormats.AppendString( mediaFormat );
    }
  }
}


void MediaServer::ApplyMediaInfo()
{
  PStringList mediaFormatOrder, mediaFormatMask;

  m_MediaInfo.sort();

  for (MediaInfo::iterator mm = m_MediaInfo.begin(); mm != m_MediaInfo.end(); ++mm) {
    if (mm->preferenceOrder < 0)
    {
      mediaFormatMask.AppendString(mm->mediaFormat);
    }else
    {
      mediaFormatOrder.AppendString(mm->mediaFormat);
      
    }
  }

  if (!mediaFormatOrder.IsEmpty()) {
    SetMediaFormatOrder(mediaFormatOrder);
    SetMediaFormatMask(mediaFormatMask);
  }
}

/////////////////////////////////////////////////////////////////



MediaServerEndPoint::MediaServerEndPoint(
  MediaServer & manager
) : OpalIVREndPoint( manager )
{

  defaultVXML = "<?xml version=\"1.0\"?>"
                "<vxml version=\"1.0\">"
                  "<form id=\"root\">"
                    "<audio src=\"welcome.wav\">"
                      "Welcome to the auto attendant main menu."
                    "</audio>"
                  "</form>"
                "</vxml>";



}

OpalIVRConnection * MediaServerEndPoint::CreateConnection(
  OpalCall & call,
  const PString & token,
  void * userData,
  const PString & vxml
)
{
  return dynamic_cast<MediaServer &>(GetManager()).CreateConnection( call, token, userData, vxml ); 
}



